#include "AST.h"
#include "CodeContext.h"
#include "CtReference.h"
#include "PrimitiveDecl.h"
#include "code-assign.h"
#include "code-util.h"
#include "code-foreach.h"
#include "ctBox.h"
#include "lgMacro.h"


const string ArrayAccessNode::cArrayHandler( const char [] )
{
  undefined( "cArrayHandler" );
  return "";
}


/*
 * Java array handlers are of the form:
 *
 *    JAVA_ARRAY_op_dist
 *
 * Where "op" is the operation being performed and "dist" is the
 * reference distance, either "LOCAL" or "GLOBAL".
 */

const string JavaArrayAccessNode::cArrayHandler( const char operation[] )
{
  return "JAVA_ARRAY_" + lgMacro( operation, array()->type()->isLocal() );
}


/*
 * Titanium array handlers are of the form:
 *
 *    TI_ARRAY_op_dist( elem, arity )
 *
 * Where:
 *
 *    - "op" is the operation being performed
 *    - "dist" is the reference distance (either "LOCAL" or "GLOBAL")
 *    - "elem" is the element type name
 *    - "arity" is the array arity
 */

const string TitaniumArrayAccessNode::cArrayHandler( const char operation[] )
{
  const TypeNode &arrayType = *array()->type();
  
  return "TI_ARRAY_" + lgMacro( operation, arrayType.isLocal() )
    + '('
    + arrayType.elementType()->cType()
    + ", "
    + int2string( arrayType.tiArity() )
    + ')';
}

// Assume all strength-reduced accesses are to titanium arrays.
const string SRArrayAccessNode::cArrayHandler( const char operation[] )
{
  const TypeNode &arrayType = *array()->type();
  
  return "TI_ARRAY_" + lgMacro( operation, arrayType.isLocal() )
    + '('
    + arrayType.elementType()->cType()
    + ", "
    + int2string( arrayType.tiArity() )
    + ')';  
}

// Assume all offset strength-reduced accesses are to titanium arrays.
const string OSRArrayAccessNode::cArrayHandler( const char operation[] )
{
  const TypeNode &arrayType = *array()->type();
  
  return "TI_ARRAY_" + lgMacro( operation, arrayType.isLocal() )
    + '('
    + arrayType.elementType()->cType()
    + ", "
    + int2string( arrayType.tiArity() )
    + ')';  
}

/**********************************************************************/


bool ArrayAccessNode::isLocalLvalue()
{
  return array()->type()->isLocal();
}

bool ArrayAccessNode::isLocalInferredLvalue()
{
  return array()->type()->isLocalInferred();
}


/*
 * We want the address of the n'th element of an array.  We may have
 * two possible ways of computing this.
 *
 * First, we can always compute it using the appropriate "ADDR"
 * handler.  This handler will typically use some manner of standard
 * array math (base + size * offset).
 *
 * Second, if the array access has been strength-reduced, the element
 * address is already available to us in strengthReducedValue().
 *
 * So.  If the array has not been strength-reduced, we use the first
 * method.  If the array has been strength-reduced, we use the second
 * method.  If the array has been strength-reduced and we are testing
 * the strength reduction code, we use both methods and compare the
 * results.
 */

/* Return the name of a C variable that points to appropriate array element. */
static const string OSRaddr(CodeContext &context,
			    OSRArrayAccessNode *t)
{
  extern int SRTempNum;
  const TypeNode *ty = t->array()->type();
  const CtType &eltType = ty->elementType()->cType();
  bool isLocal = ty->isLocal();
  const string result =
    t->offsetString() + 'Z' + int2string(unique2((void *) SRTempNum,
						 (void *) t));
  context.declare(result, ctBox(eltType, isLocal));
  context << lgMacro("SUM", isLocal) <<
    "(" << eltType << " *, " << result << ", " <<
    t->codeString() << ", " <<
    t->offsetString() << ");" << endCline;
  return result;
}

const string ArrayAccessNode::getLvalue( CodeContext &context )
{
  const bool isReduced = isStrengthReduced();
  const bool isLocal = array()->type()->isLocal();
  string standard, reduced, result;

  // Compute the standard element address
  if (test_sr || !isReduced) {
    const CtType &elementType = array()->type()->elementType()->cType();
    const CtReference &elementAddrType = ctBox( elementType,
						isLocalLvalue() );
    const string arrayTemp = array()->simpleVar( context );
    const string indexTemp = index()->emitExpression( context );
    
    result = standard = declareTemporary( context, elementAddrType );
    context << cArrayHandler( "ADDR" ) << '(' << result << ", "
	    << arrayTemp << ", " << indexTemp << ");" << endCline;
  }

  // Compute the strength-reduced element address
  if (isReduced)
    result = reduced =
      isSRArrayAccessNode(this) ?
        static_cast<SRArrayAccessNode *>(this)->codeString() :
        OSRaddr(context, static_cast<OSRArrayAccessNode *>(this));

  // They both better be the same
  if (test_sr && isReduced)
    context << "assert(" <<
      lgMacro("EQUAL", isLocal) << "(" << standard << ", " <<
      reduced << "));" << endCline;
  
  return result;
}


const string JavaArrayAccessNode::getLvalue( CodeContext &context )
{
  // temporary used to hold address of "length" field
  context.depend( ctBox(PrimitiveDecl::IntDecl.cType(),
			array()->type()->isLocal() ) );

  // offsets of "length" and "data" fields
  context.depend( array()->type()->decl()->cType() );
  
  return ArrayAccessNode::getLvalue( context );
}


/**********************************************************************/


/* Output code equivalent to "dest = *address".  The resulting code
   should be executed for side effects only.  local indicates whether
   address is a local pointer. */
string load(const string &dest, const string &address, bool local)
{
  return lgMacro("DEREF", local) + '(' + dest + ", " + address + ")";
}

const string ArrayAccessNode::emitExpression2( CodeContext &context )
{
  const TypeNode &arrayType = *array()->type();
  const bool arrayLocal = arrayType.isLocal();
  const string address = getLvalue( context );

  const TypeNode &elemType = *arrayType.elementType();
  const string elemValue = declareTemporary( context, elemType.cType() );

  context << load(elemValue, address, arrayLocal) << ';' << endCline;

  if (!arrayLocal && elemType.isLocal()) {
    // when accessing a global array of local elements, the
    // resulting element value must be widened from local to global
    const string proc = "TO_PROC(" + address + ')';
    return elemType.emitGlobalize( context, proc, elemValue );
  } else
    return elemValue;
}


const string ArrayAccessNode::receiveAssign( const string &lvalue,
					     const string &value,
					     AssignKind kind )
{
  return makeAssign( lvalue, value, array()->type()->isLocal(), kind, false );
}


/**********************************************************************/


const string ArrayInitNode::emitExpression2( CodeContext &context )
{
  vector< string > elements;
  initializers()->emitExpressionList( context, elements );

  const int length = elements.size();
  const string array = declareTemporary( context );
  context << "JAVA_ARRAY_ALLOC(" << array << ", NULL, " << length << ", "
	  << type()->elementType()->cType() << ", " << type()->elementType()->isAtomic() << ");" << endCline;
    
  for (unsigned fill = 0; fill < elements.size(); ++fill) {
    context << array << "->data[" << fill << "] = " << elements[fill] << ';' << endCline;
  }
  
  context.depend( type()->decl()->cType() );

  return array;
}

/////////////////////////////////////////////////////////////////////////////

bool ArrayAccessNode::isStrengthReduced() const
{
  return false;
}

bool SRArrayAccessNode::isStrengthReduced() const
{
  return true;
}

bool OSRArrayAccessNode::isStrengthReduced() const
{
  return true;
}

