/* code-foreach.cc */

/*

  To get good performance on foreach over RectDomains, we need to
  strength reduce foreach ourselves because of the division in the
  index expression. RectDomains can have non-unit stride, so the
  division must be there in a naive implementation.  C compilers then
  refuse to do strength reduction on the address calculation even when
  the addresses are in an arithmetic progression.

  Idea:

  Before code gen, identify index expressions that are candidates for
  strenth reduction.  E.g., A[p] where p is the iteration variable of
  a foreach.  Mark the index expression and the foreach.

  At code gen of a foreach, check every item on the list of possibly
  strength-reducible index expressions.  If the expression is simply
  A[p] then we need A to be loop invariant; if the expression is
  A[j*p+k] then we need A and j and k to be loop invariant; etc.
  There are further complications in checking validity, though.  The
  problem is that the difference between &A[p0] and &A[p1] may not be
  integral for adjacent iterations p0 and p1.  This only comes up when
  A[p] is not used on every iteration.  See the bottom of the file for
  a copy of a relevent email from May, 1997.

  If it is indeed reducible, we create a pointer and a
  pointer-delta-per-iteration for it in the code output.

  If all uses of the iteration variable are optimized, then we may omit
  the code to create the point object.

  All of the above is applied to the rects comprising a general
  Domain if we are iterating over a general Domain.

*/

#include <cassert>
#include <fstream>
#include <iomanip>
#include <list>
#include <map>
#include <set>
#include <stack>
#include "AST.h"
#include "CodeContext.h"
#include "CtGlobal.h"
#include "CtLocal.h"
#include "CtReference.h"
#include "CtType.h"
#include "PrimitiveDecl.h"
#include "code-assign.h"
#include "code-call.h"
#include "code-defs.h"
#include "code-foreach.h"
#include "code-grid.h"
#include "code-util.h"
#include "code.h"
#include "compiler.h"
#include "config.h"
#include "ctBox.h"
#include "decls.h"
#include "delimit.h"
#include "domain-decls.h"
#include "errors.h"
#include "interface.h"
#include "lgMacro.h"
#include "osstream.h"
#include "utils.h"
#include "optimize.h"
#include "code-MIVE.h"
#include "buildexpr.h"
#include "ArrayAccessSet.h"
#include "PolyToCode.h"
#include "stats.h"

#ifdef HAVE_BINARY_ADAPTORS
typedef stack< string, list< string > > StringStack;
#else
typedef stack<         list< string > > StringStack;
#endif

/* Temporary counters, etc. (should be reset once per file output) */
int SRTempNum = 0;
static map_tree_to_cMIVElist *t_memo;
static map_cMIVE_to_int *t_memo2;

static void resetTempCounters(int *c)
{
  *c = 0;
  if (t_memo)
    delete t_memo;
  if (t_memo2)
    delete t_memo2;
  t_memo = new map_tree_to_cMIVElist;
  t_memo2 = new map_cMIVE_to_int;
}

/* If we've been called before with the same t and a MIVE equal to m then
   return the same answer.  Otherwise return an answer different from any
   we've returned lately.  (Where "lately" means within the compilation of
   the current file.) */
static int tempNum(TreeNode *t, const MIVE *m)
{
  llist<const MIVE *> * & l = (*t_memo)[t];
  int &result = (*t_memo2)[m->find(l)];
  if (result == 0)
    result = ++SRTempNum;

  return result;
}

const string pointField(const string p, const string field)
{
  return p + "." + field;
}

const string pointField(const string p, int field)
{
  return p + ".x" + int2string(field);
}

const string rectField(const string r, const string field, int arity)
{
  return r + "." +
    MANGLE_TI_DOMAINS_RECTDOMAIN_FIELD_ACCESS(+, field, int2string(arity));
}

static StringStack pointGuts(CodeContext& context, TreeNode *pt)
{
  assert(!isPointNode(pt));

  StringStack q;
  int arity = pt->type()->tiArity();

  string index0 = pt->simpleVar(context);
  for (int i = 0; i < arity; i++)
    q.push(index0 + ".x" + int2string(i));

  return q;
}

static const string indexCurrent(const string &q, int i)
{
  return q + "_" + int2string(i);
}

static const string indexStart(const string &q, int i)
{
  return q + "s_" + int2string(i);
}

static const string indexStride(const string &q, int i)
{
  return q + "d_" + int2string(i);
}

static const string indexEnd(const string &q, int i)
{
  return q + "e_" + int2string(i);
}

/*
static StringStack currentPoint(const string &q, int arity)
{
  StringStack s;
  for (int i = 0; i < arity; i++)
    s.push(indexCurrent(q, i));

  return s;
}
*/

/* i is a 1-based index.  l is a foreach loop. Return a string for the
   current value of the iteration point in dimension i. */
string *& iterationPoint(TreeNode *l, int i)
{
  static map_tree_int_to_pstring m;
  return m[l][i];
}

static void
foreachRectDomainStart(CodeContext &context, int arity, const string &q,
		       TreeNode *dom, string &curr_rect)
{
  // RectDomain specific startup
  if (isObjectNode(dom) || isOFAN(dom) || isTypeFAN(dom)) {
    curr_rect = dom->emitExpression(context);
  } else if (isGridMethod_domain(dom)) {
    curr_rect = dom->child(0)->child(0)->emitExpression(context) + ".domain";
  } else {
    string exp = dom->emitExpression(context);
    curr_rect = q + "_curr_rect";
    context.declare(curr_rect, RectDomainNDecl[arity - 1]->cType());
    context << curr_rect << " = " << exp << ";" << endCline;
  }
}


/* BUG: this code should use context.declare(...) to declare variables. */
static void
foreachGeneralStart(CodeContext &context, int arity, const string &q,
		    TreeNode *dom)
{
  // General domain specific startup
  // iter = d.getRectangles().newIterator()
  // do curr_rect = iter[i]

  ClassDecl &d = *DomainNDecl[ arity - 1 ];
  ClassDecl &rdl = *RectDomainListNDecl[ arity - 1 ];
  ClassDecl &iter = *RectDomainListIteratorNDecl[ arity - 1 ];

  const CtType &d_desc = d.cDescriptorType();
  const CtType &rdl_desc = rdl.cDescriptorType();
  const CtType &iter_desc = iter.cDescriptorType();  

  const CtGlobal &gl_d_desc = *new CtGlobal( *new CtLocal( d_desc ) );
  const CtGlobal &gl_rdl_desc = *new CtGlobal( *new CtLocal( rdl_desc ) );
  const CtGlobal &gl_iter_desc = *new CtGlobal( *new CtLocal( iter_desc ) );

  const CtGlobal &d_inst = *new CtGlobal( d.cType() );
  const CtGlobal &rdl_inst = *new CtGlobal( rdl.cType() );
  const CtGlobal &iter_inst = *new CtGlobal( iter.cType() );

  const string d_var = q + "_d";
  const string mdl_var = q + "_mdl";
  const string d_iter_var = q + "_d_iter";
  const string iter_cont_var = q + "_iter_cont";
  const string d_desc_var = q + "_d_desc";
  const string mdl_desc_var = q + "_mdl_desc";
  const string d_iter_desc_var = q + "_d_iter_desc";
  const string curr_rect_var = q + "_curr_rect";

  context.depend( d_desc );
  context.depend( rdl_desc );
  context.depend( iter_desc );

  context.depend( gl_d_desc );
  context.depend( gl_rdl_desc );
  context.depend( gl_iter_desc );

  context.declare( d_var, d_inst );
  context.declare( mdl_var, rdl_inst );
  context.declare( d_iter_var, iter_inst );
  context.declare( iter_cont_var, PrimitiveDecl::BoolDecl.cType() );

  context << d_desc    << " *" << d_desc_var << ';'      << endCline
	  << rdl_desc  << " *" << mdl_desc_var << ';'    << endCline
	  << iter_desc << " *" << d_iter_desc_var << ';' << endCline;
    
  const string exp = dom->emitExpression(context);

  context << d_var << " = " << exp << ';' << endCline
	  << "CLASS_INFO_GLOBAL(" << d_desc_var << ", "
	  << d_desc << ", " << d_var << ");" << endCline
	
	  << mdl_var << " = " << d_desc_var << "->" << MANGLE_TI_DOMAINS_DOMAIN_GETRECTANGLES << d_var << ");" << endCline
	  << "CLASS_INFO_GLOBAL(" << mdl_desc_var << ", "
	  << rdl_desc << ", " << mdl_var << ");" << endCline
	
	  << d_iter_var << " = " << mdl_desc_var << "->" << MANGLE_TI_DOMAINS_RDL_NEWITERATOR << mdl_var << ");" << endCline
	  << "CLASS_INFO_GLOBAL(" << d_iter_desc_var << ", "
	  << iter_desc << ", " << d_iter_var << ");" << endCline

	  << "do {" << endCline
	  << "  " << iter_cont_var << " = !(" << d_iter_desc_var << "->" << MANGLE_TI_DOMAINS_RDL_ISEND << d_iter_var << "));" << endCline
	  << "  if (" << iter_cont_var << ") {" << endCline
	  << "     RD" << arity << " " << curr_rect_var << ";" << endCline
	  << "     " << curr_rect_var << " = " << d_iter_desc_var << "->" << MANGLE_TI_DOMAINS_RDL_DEREFERENCE << d_iter_var << ");" << endCline;
}

static void updateIterationPoint(CodeContext &os, const string &q,
				 const string iterationVar, int arity)
{
  os << MANGLE_TI_DOMAINS_POINT_CONSTRUCT(<<, int2string(arity)) <<
    MANGLE_STACK_VAR(<<, iterationVar);
  for (int i = 0; i < arity; i++)
    os << ", " << indexCurrent(q, i);
  os << ");" << endCline;
}

// Set the point-valued variable to the values in indices.
static void updatePoint(CodeContext &os, const string &v,
			StringStack indices)
{
  int arity = indices.size();
  os << MANGLE_TI_DOMAINS_POINT_CONSTRUCT(<<, int2string(arity)) << v;
  for (int i = 0; i < arity; i++, indices.pop())
    os << ", " << indices.top();
  os << ");" << endCline;
}

static string quickconv(StringStack indices, const string& arr,
			CodeContext &context)
{
  string s;
  for (int i = indices.size(); i >= 1; i--, indices.pop()) {
    string d = difference(indices.top(),
			  arr + ".base[" + int2string(i - 1) + "]", context);
    string q = quotient(d, arr + ".stride[" + int2string(i - 1) + "]",
			context);
    string p = product(q, arr + ".sideFactors[" + int2string(i - 1) + "]",
		       context);
    s = (s == "" ? p : sum(s, p, context));
  }
  return s;
}

#if 0
static string slowconv(StringStack indices, const string& array,
		       const string &q, TypeNode &t, const string &where,
		       CodeContext &o)
{
  const int arity = indices.size();
  string x = "x" + q;
  o.declare(x, makePointType(arity)->cType());
  o << MANGLE_TI_DOMAINS_POINT_CONSTRUCT(<<, int2string(arity)) <<
    MANGLE_STACK_VAR(<<, iterationVar);
  for (int i = indices.size(); i >= 1; i--, indices.pop())
    o << ", " << indices.top();
  o << ");" << endCline;
  return safePointer(array, x, t, where);
}
#endif

// Compute how far in memory we move (in units of sizeof(array elt))
// when the index of an array index changes by d times multiplier in
// the i'th dimension.  (i counts from 0.)  If bounds checking is on
// AND requireCheck is true then check that any divisions done come
// out evenly (i.e., if doing x/y then make sure y divides x).
static string convertDiffToOffset(const Poly *d, int i, TreeNode *WRTloop,
				  const string &arr,
				  const string &multiplier,
				  const string &requireCheck,
				  CodeContext &context)
{
  string result;
  if (d->zerop())
    result = "0";
  else if (d->constantp())
    result = int2string(d->asInt());
  else
    result = PolyToCode(d, WRTloop, context);

  if (result != "0") {
    current_context = &context;
    result = product(result, multiplier);
    result = quotient_with_check(result,
				 arr + ".stride[" + int2string(i) + "]",
				 requireCheck);
    result = product(result, arr + ".sideFactors[" + int2string(i) + "]");
  }
  return result;
}

// i and j are 0-based indices
static string computeDelta(const MIVE *m, int i, TreeNode *WRTloop, int j,
			   const string &arr, const string &domainStride,
			   const string &requireCheck, CodeContext &context)
{
  const Poly *d = deriv(m, i, WRTloop, j);
  assert(d != NULL);

  string result = convertDiffToOffset(d, i, WRTloop, arr, domainStride,
				      requireCheck, context);
  if (DEBUG_SR)
    cout << "compute change in " <<
      stringify(m->p[i]) << " as " << 
      PCODE(WRTloop->vars()->child(0)->simpName()) << "[" << j+1 << "]"
      " changes: " << result << " (deriv is " << stringify(d) << ")" << endl;

  return result;
}

/* Return C code for evaluating ARRAY[POINT] for a titanium array
   and a point of the appropriate arity.  

   Unless the C code is compiled with -DBOUNDS_CHECKING=0 then, at run
   time, the program will verify that the array's domain contains the
   point.  (If not, the program halts with an error message that includes
   the WHERE string.)
*/
static string safePointer(const string &array, const string &point,
			  TypeNode &t, const string &where)
{
  return callGridMethod(t, "_PTR",
			"(\"" + where + "\", "+ array + ", " + point + ")");
}

static void setupConstantPointer(TreeNode *arr, TreeNode *ind,
				 ArrayAccessSet *a,
				 TreeNode *WRTloop, CodeContext &context)
{
  TreeNode *acc = a->use(arr, ind);
  string ptr = LIPTempname(WRTloop, arr, ind);
  TypeNode *t = arr->dtype();
  bool isGlobal = !(t->modifiers() & Common::Local);
  const CtType &eltType = t->elementType()->cType();
  context.declare(ptr, ctBox(eltType, !isGlobal));
  const string array = acc->array()->emitExpression(context);
  const string index = acc->index()->emitExpression(context);
  const string p = string("loop at ") + WRTloop->position().asString();
  context << ptr << " = " <<
    safePointer(array, index,  *t, p) << ';' << endCline;
}

#define START(ns) (string("s") + (ns))
#define DIFF(ns, dim) (string("d") + (ns) + '_' + int2string(dim))
#define nonTrivialSize(dim)			\
    ('(' + indexCurrent(q, dim) + " + " +	\
     indexStride(q, dim) + " < " +		\
     indexEnd(q, dim) + ')')
// #define TrivialSize(dim) (string("(!(") + nonTrivialSize(dim) + "))")

void setupOSR(TreeNode *arr, const MIVE *base, const MIVE *dest,
	      ArrayAccessSet *a, TreeNode *WRTloop, CodeContext &context)
	      
{
  const string array = arr->emitExpression(context);
  const CtType &eltType = arr->type()->elementType()->cType();
  string result = "0";
  // m, the thing we're computing, is dest - base.
  const MIVE *m = dest->add(base->negate());
  // convert m to distance in memory
  for (int i = 0; i < m->arity(); i++) {
    string s = convertDiffToOffset(m->p[i], i, WRTloop, array, "1",
				   "1", context);
    result = sum(result, s, context);
  }
  string t = a->OSRTempname(WRTloop, arr, dest);
  context.declare(t, theIntType->cType());
  result = product(result, string("sizeof(") + eltType + string(")"), context);
  context << t << " = " << result << ';' << endCline;
}


/*
  CONTEXT is the CodeContext for outputting code.
  M contains the <array, MIVE> pairs that we care about for strength reduction.
  ARRAYACCESSES are the optimized access in the loop body WRT to this loop.
  L is this loop (ForEachStmtNode).
  Q is a string for uniquely identifying a family of C variables.
  DIM is the dimension to work on---this fn is called once per dimension.
  ARITY is the arity of the rectangle.
  ITERATIONVAR is the iteration variable.
  CURR_RECT is a string for the C variable over which we're iterating now.
  INCR is modified (by appending) to contain code needed to increment
    the pointers that are used by strength reduction.
*/
static void SRInit(CodeContext &context, map_tree_to_cMIVElist &m,
		   ArrayAccessSet *arrayAccesses, TreeNode *l, 
		   const string &q, int dim, int arity,
		   const string &iterationVar, const string &curr_rect,
		   string &incr)
{
  if (arrayAccesses == NULL) return;

  bool noneHandled = true;

  /* Part 1: Compute initial value for the SR pointers. */

  for (map_tree_to_cMIVElist::const_iterator a = m.begin(); a != m.end(); a++) {
    TreeNode *acc = arrayAccesses->use((*a).first);
    TreeNode *arr = acc->array();
    const string array = arr->emitExpression(context);
    
    foreach (z, llist<const MIVE *>, *((*a).second)) {
      const MIVE *p = *z;

      TypeNode *t = arr->type();
      bool isGlobal = !(t->modifiers() & Common::Local);
      const CtType &eltType = t->elementType()->cType();
      const int arrayArity = t->tiArity();
      const string arityStr = int2string(arrayArity);
      const string R = arrayMethodPrefix(t) + "domain("
	  + eltType + ", " + arityStr + ")(" + array + ")";
      const string ns =
	SRTempname(l, (llist<const MIVE *> * &) (*a).second, acc, p);
      const string prevbase = SRvar_at_depth(ns, dim - 1);
      const string base = SRvar_at_depth(ns, dim);
  
      /* cout << "Iteration Arity = " << arity << ";"
	 " Array Arity = " << arrayArity << endl; */
      assert(p->arity() == arrayArity);

      noneHandled = false;

      // Make a note, for later, of how to increment the pointer when
      // moving in this direction.
      incr += lgMacro("INDEX", t->isLocal()) +
	"(" + base + ", " + base + ", " + DIFF(ns, dim) + "); ";

      if (dim == 0) {
	context.declare(base, ctBox(eltType, !isGlobal));
	StringStack s;
	for (int i = 0; i < arrayArity; i++) {
	  s.push(PolyToCode(p->p[i], l, context));
	}
	string qc = quickconv(s, array, context);
	// context << START(ns) << " = " << qc << ';' << endCline;
	context << lgMacro("INDEX", t->isLocal()) << '('
		<< base << ", " << array << ".A, "
		<< qc << ");" << endCline;

	context.pushNoLineDirectives();
	context << "#if BOUNDS_CHECKING" << endCline;
	{
	  // Warning: assumes that end - 1 is the last point in a rectdomain,
	  // regardless of its stride.  This echoes the implementation of 
	  // getUpperMaxPoint() in tiRectDomainN.java.

	  // The generated code uses the iteration point as a temporary,
	  // and resets it at the end of the block.
	  string ** originalIP = new string * [arity];
	  CodeContext subcontext(context, context);
	  const string tempAddr = base + "tmp";
	  const string tempIndex = base + "W";
	  subcontext.declare(tempIndex,
			     makePointType(arrayArity)->cType());
	  subcontext.declare(tempAddr, ctBox(eltType, !isGlobal));

	  for (int i = 0; i < arity; i++)
	    originalIP[i] = iterationPoint(l, i + 1);

	  /* Using the MIVE p, map each corner of the iteration space
	     into array space and bounds check it. */
	  for (int j = 1 << arity; j-- > 0; ) {
	    StringStack s;
	    for (int i = 0; i < arity; i++)
	      iterationPoint(l, i + 1) =
		((j & (1 << i)) ?
		 new string("(" + indexEnd(q, i) + " - 1)") : 
		 originalIP[i]);
	      
	    for (int i = arrayArity; i-- > 0; )
	      s.push(PolyToCode(p->p[i], l, subcontext));
	    updatePoint(subcontext, tempIndex, s);
	    if (j == 0)
	      subcontext << tempAddr << " = ";
	    subcontext <<
	      safePointer(array, tempIndex,
			  *t, arr->position().asString()) <<
	      ';' << endCline;
	  }

	  // This is an internal check on the compiler's correctness
	  // and could be omitted.
	  subcontext << "if (!" << lgMacro("EQUAL", t->isLocal()) <<
	    "(" << base << ", " << tempAddr << "))" << endCline <<
	    "  assert((\"internal error: "
	    "bogus strength reduction\", 0));" << endCline;

	  delete[] originalIP;
	}
	context << "#endif" << endCline;
	context.popNoLineDirectives();
      } else {
        // dim > 0
	context.declare(base, ctBox(eltType, !isGlobal));
	context << base << " = " << prevbase << ";" << endCline;
      }
    }
  }
  
  /* Part 2: Constant pointers */
  if (dim == 0) {
    context << "/* loop invar pointers */" << endCline;
    for (ArrayAccessSet::lip_iterator i = arrayAccesses->lip_begin();
	 !i.isDone();
	 i.next())
      setupConstantPointer((TreeNode *) i.array(),
			   (TreeNode *) i.index(), arrayAccesses, l, context);
    context << "/* end of loop invar pointer setup */" << endCline;
  }

  if (noneHandled) return;

  /* Part 3: OSR */
  
  if (opt_osr) {
    if (dim == 0) {
      context << "/* OSR calculations */" << endCline;
      for (ArrayAccessSet::osr_iterator i = arrayAccesses->begin();
	   !i.isDone();
	   i.next())
	setupOSR(arrayAccesses->use(i.array())->array(), i.base(), i.dest(),
		 arrayAccesses, l, context);
      context << "/* end of OSR calculations */" << endCline;
    }
  } else
    context << "/* OSR disabled */" << endCline;

  /* Part 4: incrementing the pointers */

  // Compute the (pointer) increments necessary for each dimension.
  // Output that calculation inside an "if (nontrivialsize()) { ... }".
  // If the size of the domain is trivial then we don't need the delta
  // per iteration because there will only be zero or one iterations.

  if (dim == 0) {
    for (int d = 0; d < arity; d++) {
      for (map_tree_to_cMIVElist::const_iterator a = m.begin();
	   a != m.end(); a++) {
	TreeNode *acc = arrayAccesses->use((*a).first);
	TreeNode *arr = acc->array();
	const string array = arr->emitExpression(context);
	  
	foreach (z, llist<const MIVE *>, *((*a).second)) {
	  const MIVE *p = *z;

	  const string ns =
	    SRTempname(l, (llist<const MIVE *> * &) (*a).second, acc, p);

	  context.declare(DIFF(ns, d), theIntType->cType());
	  if (use_nonTrivialSizeCheck)
	    context << "if " << nonTrivialSize(d) << " {";
	  for (int i = 0; i < p->arity(); i++) {
	    string NTS =
	      use_nonTrivialSizeCheck ? string("1") : nonTrivialSize(d);
	    string domainStride = indexStride(q, d);
	    string s = computeDelta(p, i, l, d, array, domainStride,
				    NTS, context);
	    context << DIFF(ns, d) << (i == 0 ? " = " : " += ") << s <<
	      ';' << endCline;
	  }
	  if (use_nonTrivialSizeCheck)
	    context << "}" << endCline;
	}
      }
    }
  }
}
#undef DIFF
#undef nonTrivialSize



#define middle_label(i) "middle" << SRTempNum << "_" << (i)
#define top_label(i) "top" << SRTempNum << "_" << (i)

/*
  ARITY is the arity of the rectangle.
  Q is a string for uniquely identifying a family of C variables.
  DOM is the domain of the iteration (not necessarily just this rectangle).
  VAR is the iteration variable.
  STMT is the body of the loop.
  NEEDPOINT is whether to update the point-valued iteration var explicitly.
  CURR_RECT is a string for the C variable over which we're iterating now.
  CANBEEMPTY is whether curr_rect can have 0 points in it.
  M contains the <array, MIVE> pairs that we care about for strength reduction.
  ARRAYACCESSES are the optimized access in the loop body WRT to this loop.
  L is this loop (ForEachStmtNode).
*/
static void foreachRectIterate(CodeContext &context, int arity, string q,
			       TreeNode *dom, TreeNode *var, TreeNode *stmt,
			       bool needPoint, string &curr_rect,
			       bool canBeEmpty, map_tree_to_cMIVElist &m,
			       ArrayAccessSet *arrayAccesses, TreeNode *l)
{
  /*
    Approximate idea:

    iterate over RectDomain curr_rect
    s = getLoopingStride;
    p0 = getLowerBound;
    p1 = getUpperBound;
    for (var0, p0.get(0), p1.get(0), s.get(0))
      ... for (var-n, p0.get(n), p1.get(n), s.get(n))
	do stuff
      ... end for var-n
    end for var0
  */

  int i;

  /* Ensure unique labels for gotos (if any). */
  ++SRTempNum;

  // Preliminaries

  // Declare the iteration point.
  context.declare(MANGLE_STACK_VAR(+, *(var->ident())),
		  makePointType(arity)->cType());

  // Set up variables for the current value, start, stride, and end of
  // each dimension of the iteration.  Exception: for the outermost
  // loop of the iteration we don't need to keep the start around.
  // Warning: This is relies on knowledge about the domain library
  // implementation because it pulls the rectdomain variables directly
  // from the struct.
  for (i = 0; i < arity; i++) {
    context.declare(indexCurrent(q, i), theIntType->cType());
    iterationPoint(l, i + 1) = new string(indexCurrent(q, i));
    context << indexCurrent(q, i) << " = " <<
      pointField(rectField(curr_rect, "p0", arity), i) << ';' << endCline;

    if (i > 0) {
      context.declare(indexStart(q, i), theIntType->cType());
      context << indexStart(q, i) << " = " <<
	indexCurrent(q, i) << ";" << endCline;
    }

    context.declare(indexStride(q, i), theIntType->cType());
    context << indexStride(q, i) << " = " <<
      pointField(rectField(curr_rect, "loopStride", arity), i) << ';' <<
      endCline;

    context.declare(indexEnd(q, i), theIntType->cType());
    context << indexEnd(q, i) << " = " <<
      pointField(rectField(curr_rect, "p1", arity), i) << ';' << endCline;
  }

  if (canBeEmpty) {
    // If the start and end values are all 0, this is a null domain.
    // Warning: This is relies on knowledge about the domain library
    // implementation.
    context << "if (";
    for (i = 0; i < arity; i++) {
      if (i != 0)
	context << " || ";
      context << "((" << indexCurrent(q, i) << " != 0) || "
	"(" << indexEnd(q, i) << " != 0))";
    }
    context << ") {" << endCline;
  }

  string *incr = new string [arity];

  // top of the loop
  for (i = 0; i < arity; i++) {
    SRInit(context, m, arrayAccesses, l,
	   q, i, arity, *(var->ident()), curr_rect, incr[i]);

    if (i == 0)
      context << "/* Top of loop */" << endCline;

    // if i is zero we skip this, because the FOOs_BAR variable doesn't
    // exist---the correct value is already in FOO_BAR.
    if (i > 0)
      context << indexCurrent(q, i) << " = " << indexStart(q, i) << ";" <<
	endCline;

    if (use_goto) {
      context << "goto " << middle_label(i) << ";" << endCline;
      context << top_label(i) << ": ;" << endCline;
      if (!incr[i].empty())
	context << incr[i] << ';' << endCline;
      context << middle_label(i) << ": ;" << endCline;
    } else {
      context << "do {" << endCline;
    }
  }

  // body of the loop
  if (needPoint)
    updateIterationPoint(context, q, *(var->ident()), arity);
  stmt->codeGen(context);
  if (stat_foreach)
    loopStats(stmt, dom->type(), cout);

  // bottom of the loop
  for (i = arity; --i >= 0; ) {
    if (use_goto)
      context << "if ((" << indexCurrent(q, i) << " += " <<
	indexStride(q, i) << ") < " << indexEnd(q, i) << ") "
	"goto " << top_label(i) << ";" << endCline;
    else {
      context << "if ((" << indexCurrent(q, i) << " += " <<
	indexStride(q, i) << ") >= " << indexEnd(q, i) << ") break;" <<
	endCline;
      if (!incr[i].empty())
	context << incr[i] << ';' << endCline;
      context << "} while (1);" << endCline;
    }
  }

  if (canBeEmpty)
    context << "}" << endCline;

  delete[] incr;
}

#undef middle_label
#undef top_label


static void foreachGeneralEnd(CodeContext &os, int arity, string q)
{
  // General domain specific tail
  // advance curr_rect
  // end while curr_rect
  os << "     " << q << "_d_iter_desc->" << MANGLE_TI_DOMAINS_RDL_ADVANCE
     << q << "_d_iter);" << endCline;
  os << "  }" << endCline;
  os << "} while (" << q << "_iter_cont);" << endCline;
}

void ForEachStmtNode::emitStatement( CodeContext &context )
{
  static int tempCount;
  if (SRTempNum == 0)
    resetTempCounters(&tempCount);
  string q = string("f") + int2alpha(tempCount++);
  TreeNode *var = vars()->child(0)->simpName();
  TreeNode *dom = vars()->child(0)->initExpr();
  int arity = dom->type()->tiArity();
  bool isGeneralDomain =
    dom->type()->isDomainType() && !dom->type()->isRectDomainType();

  reset_buildexpr();

  context << endCline << "/* foreach */" << endCline;

  {
    CodeContext subcontext( context );

    string curr_rect;
    if (isGeneralDomain) {
      foreachGeneralStart(subcontext, arity, q, dom);
      curr_rect = q + "_curr_rect";
    } else {
      foreachRectDomainStart(subcontext, arity, q, dom, curr_rect);
    }

    {
      CodeContext bodyContext( subcontext );
      foreachRectIterate(bodyContext, arity, q, dom, var, stmt(),
			 true, curr_rect, canBeEmpty, *SR, arrayAccesses,
			 this);

    }

    if (isGeneralDomain)
      foreachGeneralEnd(subcontext, arity, q);
  }
  context << "/* hcaerof */" << endCline;
}

/*
  Further discussion of the non-integral increment problem (see
  comment at top of file):

Date: Fri, 30 May 1997 15:17:29 -0700 (PDT)
From: Geoff Pike <pike@priam.CS.Berkeley.EDU>
To: titanium-group@cs.Berkeley.EDU
In-reply-to: <338E6139.6601283B@cs.berkeley.edu> (message from Alex Aiken on
	Thu, 29 May 1997 22:10:17 -0700)
Subject: foreach, foreach, foreach

Further analysis of foreach over RectDomains.

There are two efficiency issues: bounds checks and non-integral increments.
By deciding the bounds checking errors are ERRORS and deciding that
for efficiency we may compile under the assumption that there are no
out-of-bounds array references, much of the mess of two weeks ago is gone.

What are the cases we can handle optimally?  In short, the cases where
every index expression A[e] that appears in the loop body is known to
be valid in every iteration of the loop.  Or A[e] simply appears in
every iteration and is assumed to be valid because bounds checking is
turned off.  Or A[e] appears in every iteration and is proven to be
valid in the loop header as part of bounds checking.  Also, a few
other cases.

Notation:
    p is a point (the control point of the iteration)
    R and Q are RectDomains
    c, d, and k are loop invariant points or integers
    A and B are Titanium arrays (assume 1D for simplicity)

1a. foreach (p in R) { ... A[p] ... } where we know R <= A.domain
(or foreach (p in Q) { ... A[c*p+d] ... } where c*Q+d <= A.domain)

If you want the performance guarantee, one way to make sure you get it
is to only use A[p] for a p running over a domain known to be a subset
of A.domain.  Everything one can do with array foreach falls into this
category, because "array foreach (f in A) ... f ..." is equivalent to
"foreach (p in A.domain) ... A[p] ..."

Why is it fast?  We know A[p] is valid for every p, hence it is safe
and bounds checking is unnecessary.  Furthermore, &A[p] is integral if
A[p] is valid, so the memory increment used in strength reduction is
the difference of two valid addresses, an integer.

This is the big winner.  It is trivial to have programmers understand it.
Also, my feeling is that 99% of loops will fall in this category or 1b.
HOWEVER, THE PROGRAMMER MAY HAVE TO NUDGE THE COMPILER TO MAKE SURE IT
KNOWS WHAT IT NEEDS TO KNOW.

The things listed below are gravy (and increasing obscure).

1b. foreach (p in R) { ... A[c*p+d] ...} // c and d are loop invariant
        where A[c*p+d] occurs in all iterations

(In these examples we need not know a thing about R and A.)

Example 1:  foreach (p in R) { A[p] = 7; }
Example 2*: foreach (p in R) { if (foo()) A[p] = 7; else A[p+3] = 9; }
Example 3:  foreach (p in R) { A[k*p]; if (foo()) A[k*p] = 7; }

Intuition: Assume the program does not have bounds checking errors
(i.e., bounds checking is turned off or we checked it somehow).  If
A[p] is used in every iteration it is valid in every iteration and
non-integral memory increments are unnecessary.  Example 2* may be a
problem.  Example 3 is OK because A[k*p] is appears (and hence assumed to
be valid) in every iteration.  Note that the programmer might write
this way just to be extra sure that the compiler sees &A[k*p] as an
induction variable with integral increment.

If bounds checking is turned off this is really just like case 1a,
i.e., the compiler knows that the domain of access is a RectDomain
and assumes it is a subset of the array's domain because it assumes no
bounds checking errors.

2. foreach (p in R) { ... A[p] ...}
   where (stride of R) times (memory stride of A) times (elt size of A) is
     an integer

This is a subset of the dangerous cases that happen to be no problem
because the granularity of pointers in the hardware is often 4x or 8x
different from the granularity of addresses of interest.  Example 4:
in memory we have
    -----------------------------------------
     ...  | A[0] | A[2] | A[4] | A[6] | ...   (memory stride = 1:2)
    -----------------------------------------

    int i = 0;
    foreach (p in [lo : hi]) { if (bar(i++)) A[p] = 11.0; }

In this example, we can treat A[p] as an induction variable that is
incremented by (memory stride) * (stride of [lo : hi], namely, 1) * (elt size)
each iteration, which comes out to (elt size)/2, which is known at
compile time to be an integer.

Also, sometimes it may be possible to infer that an increment much be
integral even without exact knowledge of the strides.  Compare this
to example 3:

Example 5: foreach (p in R) { A[p] = sqrt(bar()); if (foo()) A[k*p] = 7; }

Here we know that A[p] is always valid, so it's increment is integral.
But A[k*p]'s increment is just k times A[p]'s increment, so it too is integral.

A similar case is where the memory stride of A is somehow known to be
an integer, but we know not what it is.  That's good enough, though.

-----------------------------------------------------------------------------

Now, the bad cases.

Example 6*: foreach (p in R) { A[floor(p*sin(p[2]))] = 11; }

This will compile into something at best equivalent to what C or
FORTRAN would do, and probably worse because we won't be able to
remove the division from the address calculation.

Example 2*: foreach (p in R) { if (foo()) A[p] = 7; else A[p+3] = 9; }

A[p] and A[p+3] can be treated as induction variables, but they may in
general have non-integral increments.  This is the problem pointed out
by David.  Let's simplify the code to just:

Example 7*: foreach (p in R) { if (foo()) A[p] = 7; }

and ask how to compile it.  One solution is not to bother treating
&A[p] as an induction variable.  Then we do the full address
calculation (in general including some multiplies and adds and at
least one divide) each time foo() is true.  The other solutions treat
&A[p] as an induction variable.  Then there are two relevant subcases:
   i.   We don't know whether the increment is integral
   ii.  We know it is not integral
 ( iii. We know it is integral --- see "good cases" above)

For the first case I propose we'll generate something like:
    {
	int index = ..., index_stop = ...;
	if (index < index_stop) {
	    int index_inc = ..., w, inc, when_inc;
	    if (...) { when_inc = ...; w = ...; inc = ...; } // sad
	    else { when_inc = 1; w = 0; inc = ...; }         // happy 
	    T *_a = ...;
	    do {
	        if (foo()) *_a = 7;
		index += index_inc;
	        if (!(index < index_stop)) break;
	        if (++w == when_inc) { w = 0; _a += inc; }
		// or, the last line could be replaced with:
		// if (happy || (++w == when_inc && ((w = 0), true))) _a += inc
	    } while (true);
	}
    }

In case (ii) we'll generate similar but slightly simpler code.

By treating &A[p] as an induction variable we pay a few extra cycles
per iteration instead of paying a bunch of extra cycles per use. 


Geoff

*/
