#include "AST.h"
#include "Worklist.h"
#include "optimize.h"
#include "code-util.h"

static bool assignment(TreeNode *t)
{
  return isExpressionStmtNode(t) && isAssignNode(t->child(0));
}

/* t is an AssignNode.  s is a set of assignments believed to be useless.
   If the result of t is used by a statement not in s then t is useful.
   I.e., if the result of t is used by a non-useless stmt then return false.
   Also return false if t has or may have side-effects. */
static bool useless(TreeNode *t, const treeSet *s)
{
  TreeNode *lhs = t->opnd0();
  if (DEBUG_DCE) {
    cout << "Checking utility of ";
    cout << PCODE(t);
    cout << endl;
  }

  if (!isObjectNode(lhs))
    return false;
  if (!isLocalOrParam(lhs->decl()))
    return false;
  TreeNode *rhs = t->opnd1();
  
  // Do we have to execute rhs because of possible side effects?
  // If so, return false immediately.
  if (!(isObjectNode(rhs) && isLocalOrParam(rhs->decl())) &&
      !isSideEffectFreeAndCannotFail(rhs) &&
      !isConstantOrCastOfConstant(rhs) &&
      !isNullOrCastOfNull(rhs) &&
      !isTrivialPointOrDomain(rhs))
    return false;

  treeSet *u = exprset_to_stmtset(RvalueUses(lhs));
  bool r = isSubset(u, s);
  if (DEBUG_DCE) {
    cout << "Stmts with rvalue uses stemming from ";
    cout << PLCODE(lhs);
    cout << ":" << endl;
    printTreeSet(u, cout, 1);
    cout << endl << "Therefore, ";
    cout << PCODE(t);
    cout << (r ? " is" : " is not") << " useless" << endl;
  }
  return r;
}

static void showAssignmentsAndUses(const treeSet *r, ostream &os)
{
  for (treeSet::const_iterator i = r->begin();
       i != r->end();
       i++) {
    (*i)->pseudoprint(os, 0);
    os << endl << "  Uses:" << endl;
    printTreeSetWithParents(list_to_set((*i)->child(0)->child(0)->getUses()), os, 2);
    os << endl << "  Defs:" << endl;
    printTreeSetWithParents(list_to_set((*i)->child(0)->child(0)->getDefs()), os, 2);
    os << endl;
  }
}

treeSet * TreeNode::uselessAssignments()
{
  Worklist w(this);
  w.filter(assignment);

  treeSet *r = list_to_set(w.asList());

  if (DEBUG_DCE) {
    cout << "Preparing to eliminate useless assignments." << endl << endl;
    showAssignmentsAndUses(r, cout);
    cout << endl;
  }

  bool progress;
  do {
    progress = false;
    for (treeSet::iterator i = r->begin();
	 i != r->end();
	 i++)
      if (!useless((*i)->child(0), r)) {
	r->erase(i);
	progress = true;
      }
  } while (progress);

  if (DEBUG_DCE) {
    cout << "Useless Assignments:" << endl;
    printTreeSet(r, cout, 1);
  }
  return r;
}

#if 0
static bool hasNoUses(VarDeclNode *t, treeSet *useless)
{
  TreeNode *var = t->simpName();
  treeSet *uses = list_to_set(var->getUses());
  if (DEBUG_DCE) {
    cout << "call hasNoUses(";
    cout << PCODE(t);
    cout << ")" << endl;
    cout << "uses of var: ";
    cout << PLCODE(var);
    cout << endl;
    printTreeSet(uses, cout, 2);
  }
  return false;
}

void ForEachStmtNode::checkIfIterationPointIsNeeded(treeSet *useless)
{
  TreeNode *var = vars()->child(0)->simpName();
  treeSet *uses = list_to_set(var->getUses());
  if (DEBUG_DCE) {
    cout << "Checking uses of foreach iteration point ";
    cout << PLCODE(var);
    cout << endl;
    printTreeSet(uses, cout, 2);
  }
  needPoint = !isSubset(uses, useless);
  if (DEBUG_DCE) {
    cout << "needPoint = " << needPoint << endl;
  }
}

treeSet *removeUselessVarDecls(TreeNode *t, treeSet *in, treeSet *out)
{
  if (isVarDeclNode(t) && hasNoUses((VarDeclNode *) t, in))
    return adjoin(out, t);

  if (isForEachStmtNode(t))
    ((ForEachStmtNode *) t)->checkIfIterationPointIsNeeded(in);

  for (int i = t->arity(); i-- > 0 ; )
    out = removeUselessVarDecls(t->child(i), in, out);

  return out;
}

TreeNode *removeUselessVarDecls(TreeNode *t, treeSet *useless)
{
  treeSet *x = new treeSet;
  TreeNode *result = t->deleteStmts(removeUselessVarDecls(t, useless, x));
  delete x;
  return result;
}
#endif
