/*
 * (c) Copyright 1997
 * Remzi H. Arpaci-Dusseau
 * remzi@cs.berkeley.edu
 */

#include "config.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stream.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/resource.h>
#include <sys/processor.h>
#include <sys/procset.h>

#include <main.h>
#include <gc.h>
#include <mem-smp.h>

#define safe_malloc	malloc
#define safe_free(x)    

int debug = 0;
#define DEBUG(args)  do { if (debug) printf args; } while (0)

#include <thread_ctl.h>


/* 
 * global variables 
 */ 
mutex_t				smp_libLock;
int				smp_libCount;

int				smp_procsTotal;
int				smp_procsOnline;

processorid_t			*smp_procArray;

int
Thread_num_smp_procs()
{
    return sysconf(_SC_NPROCESSORS_ONLN);
}

void
Thread_Init(int procs)
{
    int				i, p;
    processor_info_t		procInfo;

    /* init global lock */
    if (mutex_init(&smp_libLock, USYNC_THREAD, NULL) < 0) {
	perror("mutex_init");	
	exit(1);
    }
    smp_libCount = 0;

    /* find out processor info */
    smp_procsTotal  = sysconf(_SC_NPROCESSORS_CONF);
    smp_procsOnline = sysconf(_SC_NPROCESSORS_ONLN);

    if (procs > smp_procsOnline) {
	fprintf(stderr,
		"This backend cannot run %d Titanium processes with only %d\n"
		"physical processors.  If you need to simulate higher degrees of\n"
		"parallelism, please use the POSIX threads backend by adding\n"
		"`--backend pthreads' to your `tcbuild' command line.\n",
		 procs, smp_procsOnline);
	exit(EINVAL);
    }

    /* allocate processor array */
    smp_procArray = (processorid_t *) safe_malloc(smp_procsOnline * sizeof(processorid_t));

    /* must fill processor array */
    p = 0;
    procInfo.pi_state = P_OFFLINE; /* to passify purify */
    for (i = 0; p < smp_procsOnline; i++) {
	if (processor_info(i, &procInfo) == 0) {
	    if (procInfo.pi_state == P_ONLINE) {
		/* success */	
		smp_procArray[p] = i;
	    
		/* inc counters */
		p++;
	    }
	}
    }

    if (p != smp_procsOnline) {
	fprintf(stderr, "Thread_Init: WARNING only %d procs online\n", p);
	smp_procsOnline = p;
    }

    /* debug info */
    DEBUG(("Thread_Init: procs %d of %d online (array %d)\n", smp_procsOnline, smp_procsTotal, p));
    
    return;
}



ThreadBarrier_t *
Thread_BarrierInit(int goal)
{
    ThreadBarrier_t		*b;

    b = (ThreadBarrier_t * ) safe_malloc(sizeof(ThreadBarrier_t));

    b->tb_curr = 0;
    b->tb_goal = goal;
    mutex_init(&b->tb_lock, USYNC_THREAD, 0);
    b->done = 0;

    if (goal <= 0) {
	fprintf(stderr, "Thread_BarrierInit: failed\n");
	safe_free(b);
	return NULL;
    }

    return b;
}

void
Thread_BarrierFini(ThreadBarrier_t *b)
{
    safe_free(b);
}


int
Thread_Bind(int proc)
{
    int				t;

    if (proc < 0) {
	/* grab lock */
	mutex_lock(&smp_libLock);

	t = smp_libCount;
	smp_libCount++;
	if (smp_libCount > smp_procsOnline) {
	    smp_libCount = 0;
	}

	/* release lock */
	mutex_unlock(&smp_libLock);

	proc = smp_procArray[t];
	DEBUG(("Thread_Bind: %d(%d) -> %d\n", t, thr_self(), smp_procArray[t]));
    } 

    /* BIND it */
    if (processor_bind(P_LWPID, P_MYID, (processorid_t) proc, NULL) < 0) {
	perror("processor_bind");
    }

    return t;
}



ThreadControl_t *
Thread_CreateWorkers(int procs, void *(*func)(void *), void *args[], long flags)
{
    int				i;
    int				rc;

    ThreadControl_t		*t;

    /* debug info */
    DEBUG(("Thread_CreateWorkers: creating %d\n", procs));

    /* make control struct for this group of threads */
    t = (ThreadControl_t *) safe_malloc(sizeof(ThreadControl_t));

    t->tc_procs      = procs;
    t->tc_threadInfo = (thread_t *) safe_malloc(procs * sizeof(thread_t));
    t->tc_flags      = flags;

    /* create some threads */
    for (i = 0; i < procs; i++) {
	/* no args, why force the issue */
	if (args == NULL) {
	    rc = thr_create(NULL, 0, func, NULL, flags, &t->tc_threadInfo[i]);
	} else {
	    rc = thr_create(NULL, 0, func, (void *)args, flags, &t->tc_threadInfo[i]);
	}

	/* check for error */
	if (rc != 0) {
	    perror("thr_create");
	    return NULL;
	}
    }

    /* return SUCCESS */
    return t;
}

void
Thread_DeleteWorkers(ThreadControl_t *t)
{
    safe_free(t->tc_threadInfo);
    safe_free(t);

    return;
}


void
Thread_Wait(ThreadControl_t *t)
{
    int				i;

    /* wait for threads to exit */
    for (i = 0; i < t->tc_procs; i++) {
	if (thr_join(t->tc_threadInfo[i], NULL, NULL)) {
	    perror("thr_join");
	    return;
	}
    }

    /* why not call from here? */
    Thread_DeleteWorkers(t);

    return;
}

ThreadControl_t		*thr_c;
ThreadBarrier_t		*thr_b;

extern int ti_argc;
extern char **ti_argv;

void
Thread_Barrier(ThreadBarrier_t *b, int id)
{
    /* acquire lock */
    mutex_lock(&b->tb_lock);

    /* inc counter */
    b->tb_curr++;

    /* if we've reached the goal */
    if (b->tb_curr == b->tb_goal)
	b->done = 1;

    /* release lock */
    mutex_unlock(&b->tb_lock);
    
    while (!(b->done));

    mutex_lock(&b->tb_lock);
    b->tb_curr--;
    if (b->tb_curr == 0)
	b->done = 0;
    mutex_unlock(&b->tb_lock);

    while (b->done);
}

#ifndef MEMORY_DISTRIBUTED
int barrier()
{
    Thread_Barrier(thr_b, MYPROC);
}
#endif

void *
init_thread(void *unused)
{
    int my_pe;
    /* each thread should bind itself */
    my_pe = Thread_Bind(-1);
    /* -1 as arg does a round-robin binding, otherwise, binds to that proc */

    /* to do barrier, need to pass in b, as well as thread id */
    /*    Thread_Barrier(b, id);*/

    /* Set MYPROC and call titanium_main */
    ti_set_proc(my_pe);
    barrier();

    /* Run the user code */
    ti_main(ti_argc, ti_argv);
    barrier();

    /* Have all threads exit together. */
    thr_exit(0);
    return NULL;
}

void region_init(void);

void
Tit_Thread_Init()
{
    Thread_Init(PROCS);
    thr_b = Thread_BarrierInit(PROCS);
    thr_c = Thread_CreateWorkers(PROCS, init_thread, NULL, THR_BOUND);

    /* wait for ALL threads to exit */
    Thread_Wait(thr_c);

    /* clean up */    
    Thread_BarrierFini(thr_b);
}
