/* scope.cc: Definitions for member functions that manage name binding */

#include <assert.h>
#include "AST.h"
#include "FieldDecl.h"
#include "MethodDecl.h"
#include "compiler.h"
#include "decls.h"
#include "errors.h"

/* A default constructor for the class declared by ClassDeclNode CL, */
/* as it would be produced by the parser, had it been written */
/* explicitly:  [public] Foo() { super(); } */ 
static TreeNode* defaultConstructor (TreeNode* cl)
{
  return new ConstructorDeclNode
    ((Common::Modifiers)(cl->flags() & Common::Public),
     NULL,
     new NameNode(TreeNode::omitted, cl->simpName()->ident(), NULL),
     NULL,
     new SuperConstructorCallNode(NULL, NULL, cl->position()),
     TreeNode::omitted,
     cl->position());
}




/* Name resolution, pass 1: for ClassDeclNodes and */
/* InterfaceDeclNodes: modify *THIS to create Decls for all its */
/* members */
/* PACKAGE is the package environment containing all these nodes. */ 
void TreeNode::resolveClass(Decl *package, ClassDecl *cclass, Environ *fileEnv)
{
  foriter (p, allChildren(), ChildIter)
    (*p)->resolveClass(package, cclass, fileEnv);
}

void CompileUnitNode::resolveClass(Decl *, ClassDecl *, Environ *)
{
  foriter (type, types()->allChildren(), ChildIter)
    (*type)->resolveClass(thePackage, NULL, environ);
}

void ClassDeclNode::resolveClass(Decl *package, ClassDecl *, Environ *fileEnv)
{
  ClassDecl &me = *decl();

  me.visits = 1;

  assert(!(flags() & Interface));

  if (flags() & Immutable)
    {
      if (!superClass()->absent())
	Error("immutable classes cannot extend classes", position());

      if (interfaces()->arity() > 0)
	Error("immutable classes cannot implement interfaces", position());

      flags((Modifiers)(flags() | Final));
    }
  else
    {
      ClassDecl *super = superClass()->absent()
	? ObjectDecl
	: static_cast<ClassDecl *>(superClass()->decl());

      if (super->category() != Decl::Class)
	{
	  Error("class " + *simpName()->ident() + " cannot EXTEND interface " +
		*super->name(), position());
	  super = ObjectDecl;
	}
      superClass(super->asType());
      me.superClass(super);

      me.interfaces(NULL);
      foriter (interface, interfaces()->allChildren(), ChildIter)
	{
	  Decl *intf = (*interface)->decl();

	  if (intf->category() != Decl::Interface)
	    Error("class " + *simpName()->ident() + " cannot IMPLEMENT class " +
		  *intf->name(), position());
	  else
	    me.interfaces(cons(intf, me.interfaces()));
	}
    }

  members()->resolveClass(package, &me, fileEnv);

  if (me.environ()->lookupProper(me.name(), Decl::Constructor) == NULL)
    {
      members()->child(0, defaultConstructor(this));
      members()->child(0)->resolveClass(package, &me, fileEnv);
    }
}

void FieldDeclNode::resolveClass(Decl *package, ClassDecl *cclass, Environ *fileEnv)
{
  if (!cclass->superClass() && !(flags() & Static)) // immutable class
    flags((Common::Modifiers)(flags() | Final));

  if ((flags() & Final) && (flags() & Volatile))
    Error("field " + *simpName()->ident() + " cannot be final and volatile",
	  position());

  Decl* d;
  dtype()->resolveClass(package, cclass, fileEnv);
  d = cclass->environ()->lookupProper(simpName()->ident(), Decl::Field);
  if (d != NULL)
    {
      Error ("redeclaration of " + d->errorName(), position());
      simpName()->decl(d);
      return;
    }
  d = new FieldDecl(simpName()->ident(), dtype(), cclass, flags(), this);
  cclass->environ()->add(d);
  simpName()->decl(d);
}

void MethodDeclNode::resolveClass(Decl *package, ClassDecl *cclass, Environ *fileEnv)
{
  // methods in immutable classes and private methods are final
  if ((flags() & Private) || !cclass->superClass())
    flags((Modifiers)(flags() | Final));

  if (body()->absent())
    {
      if (flags() & Abstract)
	{
	  if (flags() & (Private | Static | Final | Synchronized | Native |
			 Inline))
	    Error("can't use private, static, final, synchronized, inline or native with abstract", position());
	}
      else if (!(flags() & Native))
	Error ("abstract or native required on methods without a body", position());
    }
  else if (flags() & (Abstract | Native))
    Error("an abstract or native method cannot have a body", position());

  MethodNode::resolveClass(package, cclass, fileEnv);
}

void MethodNode::resolveClass(Decl *, ClassDecl *cclass, Environ *)
{
  TypeNode *mType = typeFromMethodDeclNode(this);

  if ((flags() & (Local | Static)) == (Local | Static))
    Error("a static method cannot be local", position());

  // Check that this method is legal
  foriter (dd, cclass->environ()->lookupFirstProper(simpName()->ident(), Decl::Method),
	   EnvironIter)
    if (mType->methodsConflict(dd->type()))
      Error("illegal overloading of " + dd->errorName(), position());

  Decl* d = new MethodDecl(simpName()->ident(), mType, Decl::Method,
			   cclass, flags(), this);
  cclass->environ()->add(d);
  simpName()->decl(d);
}

void ConstructorDeclNode::resolveClass(Decl *, ClassDecl *cclass, Environ *)
{
  TypeNode *mType = typeFromMethodDeclNode(this);

  // Check that this method is legal
  foriter (dd, cclass->environ()->lookupFirstProper(simpName()->ident(),
						   Decl::Constructor),
	   EnvironIter)
    if (mType->methodsConflict(dd->type()))
      Error("illegal overloading of " + dd->errorName(), position());

  Decl* d = new MethodDecl(cclass->name(), mType, Decl::Constructor,
			   cclass, flags(), this);
  cclass->environ()->add(d);
  simpName()->decl(d);
}

void StaticInitNode::resolveClass(Decl *, ClassDecl *, Environ *)
{
}

void InterfaceDeclNode::resolveClass(Decl *package, ClassDecl *, Environ *fileEnv)
{
  ClassDecl &me = *decl();

  me.visits = 1;

  me.interfaces(NULL);
  foriter (interface, interfaces()->allChildren(), ChildIter)
    {
      Decl *intf = (*interface)->decl();

      if (intf->category() != Decl::Interface)
	Error("interface " + *simpName()->ident() + " cannot EXTEND class " +
	      *intf->name(), position());
      else
	me.interfaces(cons(intf, me.interfaces()));
    }

  members()->resolveClass(package, &me, fileEnv);

  foriter (decl, me.environ()->allProperDecls(), EnvironIter)
    {
      int modifiers = decl->modifiers();
      switch (decl->category())
	{
	case Decl::Method:
	  modifiers |= Public | Abstract;
	  break;
	case Decl::Field:
	  modifiers |= Public | Final | Static;
	  break;
	}
      decl->modifiers((Modifiers) modifiers);
    }
}
