/* st-name.cc: Name resolution for simple and qualified names, resolve
   parsing ambiguities.

After this phase, all fields and methods are referred to via
ThisFieldAccessNode, SuperFieldAccessNode or ObjectFieldAccessNode. 
ObjectNode is only used for local variables and parameters.

The decl in methods may be wrong, because overloading resolution is
done later (when types become available) */

#include "AST.h"
#include "NameContext.h"
#include "decls.h"
#include "errors.h"

				/* Utilities */

/* Name resolution, pass 2: return a destructively modified */
/* version of *THIS, with all names resolved to ResolvedNodes and */
/* appropriate ObjectNodes re-written as field accesses.  ENV is */
/* the environment in which to do the resolution. */
TreeNode* TreeNode::resolveName(const NameContext *ctx)
{
  const int numChildren = arity();
  for (int sweep = 0; sweep < numChildren; ++sweep)
    child(sweep, child(sweep)->resolveName(ctx));
  
  return this;
}

TreeNode *CompileUnitNode::resolveName(const NameContext *)
{
  NameContext c;

  c.env = environ;
  c.currentPackage = thePackage;
  types()->resolveName(&c);
  return this;
}

TreeNode* ClassDeclNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);

  subCtx.currentClass = decl()->asType();
  subCtx.env = decl()->environ();
  members()->resolveName(&subCtx);
  return this;
}

TreeNode* FieldDeclNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);

#if 0
  // Not true any more (late final initialisation allowed)
  if ((decl()->modifiers() & Final) && initExpr()->absent())
    Error("initialiser omitted on final " + decl()->errorName(), position());
#endif
  // field initialisers are part of the constructor
  subCtx.thisLocal = ctx->currentClass->kind() != ImmutableKind;

  dtype()->resolveName(ctx);
  initExpr(initExpr()->resolveName(&subCtx));
  return this;
}

TreeNode* VarDeclNode::resolveName(const NameContext *ctx)
{
  dtype()->resolveName(ctx);
  initExpr(initExpr()->resolveName(ctx));
  Decl *other = ctx->env->lookup(simpName()->ident(), Decl::Formal);
  if (other != NULL)
    Error("declaration shadows " + other->errorName(), position());
  else if ((other = ctx->env->lookupProper(simpName()->ident(),
					    Decl::LocalVar)) != NULL)
    Error("redeclaration of " + other->errorName(), position());
  else {
    Decl* d = new LocalVarDecl(simpName()->ident(),
				dtype(), this);
    ctx->env->add(d);
    simpName()->decl(d);
  }
  return this;
}

TreeNode* MethodDeclNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);

  subCtx.inLoop = subCtx.breakOK = NULL;
  subCtx.thisLocal = (flags() & Local) != 0
    && ctx->currentClass->kind() != ImmutableKind;
  Environ newEnv1(ctx->env);
  subCtx.env = &newEnv1;
  params()->resolveName(&subCtx);
  overlaps()->resolveName(&subCtx);
  returnType()->resolveName(ctx);
  throws()->resolveName(ctx);
  Environ newEnv2(&newEnv1);
  subCtx.env = &newEnv2;
  body(body()->resolveName(&subCtx));

  return this;
}

static void resolveOverlapName(const TreeNode::NameContext *ctx, TreeNode *name)
{
  Decl *d0 = ctx->env->lookupProper(name->ident(), Decl::Formal);

  if (!d0)
    Error("unknown parameter " + *name->ident() + " in overlap() declaration.",
	  name->position());
  name->decl(d0);
}

TreeNode* OverlapNode::resolveName(const NameContext *ctx)
{
  resolveOverlapName(ctx, opnd0());
  resolveOverlapName(ctx, opnd1());
  return this;
}

TreeNode* MethodSignatureNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);

  Environ newEnv1(ctx->env);
  subCtx.env = &newEnv1;
  params()->resolveName(&subCtx);
  returnType()->resolveName(ctx);
  throws()->resolveName(ctx);

  return this;
}

TreeNode* ConstructorDeclNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);

  subCtx.inLoop = subCtx.breakOK = NULL;
  subCtx.thisLocal = ctx->currentClass->kind() != ImmutableKind;    
  Environ newEnv1(ctx->env);
  subCtx.env = &newEnv1;
  params()->resolveName(&subCtx);
  throws()->resolveName(ctx);
  Environ newEnv2(&newEnv1);
  subCtx.env = &newEnv2;
  constructorCall(constructorCall()->resolveName(&subCtx));
  body(body()->resolveName(&subCtx));
  return this;
}

TreeNode* InterfaceDeclNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);

  subCtx.currentClass = decl()->asType();
  subCtx.env = decl()->environ();
  members()->resolveName(&subCtx);
  return this;
}

TreeNode* BlockNode::resolveName(const NameContext *ctx)
{
  Environ newEnv(ctx->env);
  NameContext subCtx(ctx, &newEnv);
  stmts((TreeListNode *)stmts()->resolveName(&subCtx));
  return this;
}

TreeNode* LabeledStmtNode::resolveName(const NameContext *ctx)
{
  Decl *other = ctx->env->lookup(label()->ident(), Decl::StmtLabel);
  if (other != NULL)
    Error("duplicate " + other->errorName(), position());

  Environ newEnv(ctx->env);
  NameContext subCtx(ctx, &newEnv);
  Decl* d = new StmtLblDecl(label()->ident(), this);
  label()->decl(d);
  newEnv.add(d);
  stmt(stmt()->resolveName(&subCtx));
  return this;
}

TreeNode* SwitchNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);

  expr(expr()->resolveName(ctx));
  Environ newEnv(ctx->env);
  subCtx.breakOK = this;
  subCtx.env = &newEnv;
  switchBlocks()->resolveName(&subCtx);
  return this;
}

TreeNode* LoopNode::resolveName(const NameContext *ctx)
{
  test(test()->resolveName(ctx));

  Environ newEnv(ctx->env);
  NameContext subCtx(ctx, &newEnv);
  subCtx.breakOK = subCtx.inLoop = this;
  stmt(stmt()->resolveName(&subCtx));
  return this;
}

TreeNode* ForNode::resolveName(const NameContext *ctx)
{
  Environ newEnv(ctx->env);
  NameContext subCtx(ctx, &newEnv);

  init((TreeListNode *)init()->resolveName(&subCtx));
  subCtx.breakOK = subCtx.inLoop = this;
  test(test()->resolveName(&subCtx));
  update((TreeListNode *)update()->resolveName(&subCtx));
  stmt(stmt()->resolveName(&subCtx));
  return this;
}

static TreeNode* resolveJump(TreeNode* node, TreeNode *noLabel, Environ *env)
{
  if (node->label()->absent())
    node->destination(noLabel);
  else
    {
      Decl *dest = env->lookup(node->label()->ident(), Decl::StmtLabel);
      node->label()->decl(dest);
      if (dest == NULL)
	Error("label " + *node->label()->ident() + " not found", node->position());
      else
	node->destination(dest->source()->stmt());
    }
  return node;
}
    
TreeNode* BreakNode::resolveName(const NameContext *ctx)
{
  if (label()->absent() && ! ctx->breakOK)
    error("unlabeled break only allowed in loops or switches");

  resolveJump(this, ctx->breakOK, ctx->env);
  return this;
}

TreeNode* ContinueNode::resolveName(const NameContext *ctx)
{
  if (! ctx->inLoop)
    Error("unlabeled continue only allowed in loops", position());
  resolveJump(this, ctx->inLoop, ctx->env);

  if (destination() && !destination()->isLoop())
    Error("continue's target is not a loop", position());

  return this;
}

TreeNode* ParameterNode::resolveName(const NameContext *ctx)
{
  dtype()->resolveName(ctx);
  Decl *other = ctx->env->lookup(simpName()->ident(),
				  Decl::Formal | Decl::LocalVar);
  if (other != NULL)
    {
      Error("parameter shadows " + other->errorName(), position());
      return this;
    }
  Decl* d = new FormalParameterDecl(simpName()->ident(),
				     dtype(), this);
  simpName()->decl(d);
  ctx->env->add(d);
  return this;
}

TreeNode* ObjectNode::resolveName(const NameContext *ctx)
{
  if (ctx->resolveAsObject)
    return resolveAName(name(), *ctx->env, ctx->currentClass, ctx->thisLocal,
			ctx->currentPackage,
			Decl::Field | Decl::LocalVar | Decl::Formal);
  else
    return resolveAName(name(), *ctx->env, ctx->currentClass, ctx->thisLocal,
			ctx->currentPackage,
			Decl::Method);
}
			      
TreeNode* ObjectFieldAccessNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);
  subCtx.resolveAsObject = true;
  object(object()->resolveName(&subCtx));
  return this;
}

TreeNode* TypeFieldAccessNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);
  subCtx.resolveAsObject = true;
  ftype((TypeNode *)ftype()->resolveName(&subCtx));
  return this;
}

TreeNode* ThisFieldAccessNode::resolveName(const NameContext *ctx)
{
  theClass(ctx->currentClass);
  isLocal(ctx->thisLocal);
  return this;
}

TreeNode* SuperFieldAccessNode::resolveName(const NameContext *ctx)
{
  theClass(ctx->currentClass);
  isLocal(ctx->thisLocal);
  return this;
}

TreeNode* ThisNode::resolveName(const NameContext *ctx)
{
  theClass(ctx->currentClass);
  isLocal(ctx->thisLocal);
  return this;
}

TreeNode* MethodCallNode::resolveName(const NameContext *ctx)
{
  args()->resolveName(ctx);

  NameContext subCtx(ctx);
  subCtx.resolveAsObject = false;
  method(method()->resolveName(&subCtx));
  return this;
}

TreeNode* CatchNode::resolveName(const NameContext *ctx)
{
  Environ newEnv(ctx->env);
  NameContext subCtx(ctx, &newEnv);

  param(param()->resolveName(&subCtx));
  block(block()->resolveName(&subCtx));

  return this;
}

TreeNode* ForEachStmtNode::resolveName(const NameContext *ctx)
{
  Environ newEnv(ctx->env);
  NameContext subCtx(ctx, &newEnv);

  // arity cannot yet be computed (no constant folding, etc), so
  // we give the Decls a dummy type (Point<0>). This will be fixed
  // later.
  TypeNode *ptype = makePointType(0);

  foriter (var, vars()->allChildren(), TreeNode::ChildIter)
    {
      (*var)->initExpr((*var)->initExpr()->resolveName(ctx));
      const string *id = (*var)->simpName()->ident();

      Decl *other = newEnv.lookup(id, Decl::Formal);
      if (other != NULL)
	Error("declaration shadows " + other->errorName(), position());
      else
	{
	  other = newEnv.lookupProper(id, Decl::Formal | Decl::LocalVar);
	  if (other != NULL)
	    Error("redeclaration of " + other->errorName(), (*var)->position());
	  else
	    {
	      Decl* d = new LocalVarDecl(id, ptype, *var);
	      newEnv.add(d);
	      (*var)->simpName()->decl(d);
	    }
	}
    }

  subCtx.inLoop = subCtx.breakOK = this;
  stmt(stmt()->resolveName(&subCtx));

  return this;
}

TreeNode* PartitionStmtNode::resolveName(const NameContext *ctx)
{
  NameContext subCtx(ctx);
  Environ newEnv(ctx->env);

  if (simpName()->absent())
    subCtx.partitionEnv = ctx->env;
  else
    {
      Decl *other = ctx->env->lookup(simpName()->ident(), Decl::Formal);
      if (other != NULL)
	Error("declaration shadows " + other->errorName(), position());
      else
	{
	  Decl* d = new LocalVarDecl(simpName()->ident(),
				     theIntType, this);
	  newEnv.add(d);
	  simpName()->decl(d);
	  // The new variable is not in scope in the conditions
	  subCtx.partitionEnv = &newEnv;
	}
    }
  cases()->resolveName(&subCtx);
  return this;
}

TreeNode *PartitionClauseNode::resolveName(const NameContext *ctx)
{
  // The partition variable is not in scope in the condition, but is
  // in the statement...
  condition(condition()->resolveName(ctx));
  NameContext subCtx(ctx, ctx->partitionEnv);
  stmt(stmt()->resolveName(&subCtx));

  return this;
}

// Misc support methods
bool TreeNode::isLoop() const { return false; }
bool IterationNode::isLoop() const { return true; }
