// Do miscellaneous tree rewrites to keep the backend happy

#include "AST.h"
#include "st-field.h"
#include "decls.h"
#include "code-util.h"

TreeNode *TreeNode::rewrite( const FieldContext *context )
{
  const int count = arity();
  for (int replace = 0; replace < count; ++replace)
    child( replace, child( replace )->rewrite( context ) );

  return this;
}

TreeNode *CompileUnitNode::rewrite( const FieldContext * )
{
  FieldContext context;
  context.currentPackage = thePackage;
  
  return TreeNode::rewrite( &context );
}

TreeNode *ClassDeclNode::rewrite( const FieldContext *context )
{
  FieldContext subcontext( *context );
  subcontext.currentClass = decl()->asType();
  
  return TreeNode::rewrite( &subcontext );
}

TreeNode *ThisFieldAccessNode::rewrite( const FieldContext * )
{
  return new ObjectFieldAccessNode(new ThisNode(theClass(),
						isLocal(),
						position()),
				   simpName(),
				   position());
}

TreeNode *SuperFieldAccessNode::rewrite( const FieldContext * )
{
  return new ObjectFieldAccessNode(new ThisNode(theClass(),
						isLocal(),
						position()),
				   simpName(),
				   position());
}

TreeNode *ArrayAccessNode::rewrite( const FieldContext *context )
{
  const TypeNode &arrayType = *array()->type();
  assert( arrayType.isArrayType() );

  TreeNode::rewrite( context );

  if (arrayType.isJavaArrayType())
    return new JavaArrayAccessNode( array(), index(), position() );
  else if (arrayType.isTitaniumArrayType())
    return new TitaniumArrayAccessNode( array(), index(), position() );
  else if (arrayType.isPointType())
    return new PointArrayAccessNode( array(), index(), position() );
  else
    {
      fatal( "array access unexpectedly applied to " + arrayType.errorName() );
      return this;
    }
}

TreeNode *VarDeclNode::rewrite( const FieldContext *context )
{
  if (initExpr()->absent()) {
    if (dtype()->isTitaniumArrayType() || dtype()->isImmutable()) {
      // handled in VarDeclNode::codeGen()
    } else
      initExpr(dtype()->defaultInitExpr(position()));
  } else
    initExpr(initExpr()->rewrite( context ));

  return this;
}

TreeNode *StringConcatNode::rewrite( const FieldContext *context )
{
  TreeNode::rewrite( context );
  return buildConcatTree( 0, context )->rewrite( context );
}


TreeNode *StringConcatAssignNode::rewrite( const FieldContext *context )
{
  // ??: The left-hand side is evaluated twice.  This will generate
  // incorrect results if it is an expression with side effects.
  
  TreeNode * const appends  = buildConcatTree( cons( opnd0() ), context );
  AssignNode * const assign = new AssignNode( opnd0(), appends );
  
  return assign->rewrite( context );
}


TreeNode *SynchronizedNode::rewrite( const FieldContext *context )
{
  const SourcePosn &pos = position();
  
  MonitorFetchNode * const fetch = expr()->absent()
    ? (MonitorFetchNode *)new MonitorFetchClassNode( context->currentClass->decl(), pos )
    : (MonitorFetchNode *)new MonitorFetchInstanceNode( expr(), pos );

  MonitorLockNode   * const lock   = new MonitorLockNode(   fetch, pos );
  MonitorUnlockNode * const unlock = new MonitorUnlockNode( fetch, pos );

  FinallyNode * const finally = new FinallyNode( unlock, pos );
  TryStmtNode * const attempt = new TryStmtNode( new TryNode( stmt(), pos ),
						 0, finally, pos );

  llist< TreeNode * > *sequence = 0;
  push( sequence, static_cast< TreeNode * >( attempt ) );
  push( sequence, static_cast< TreeNode * >( lock ) );
  push( sequence, static_cast< TreeNode * >( fetch ) );
  
  BlockNode * const block = new BlockNode( sequence, pos );
  return block->rewrite( context );
}
